/*
= StarPU-Top for StarPU =

Copyright (C) 2011 
William Braik
Yann Courtois
Jean-Marie Couteyen
Anthony Roy

This library is free software; you can redistribute it and/or
modify it under the terms of the GNU Lesser General Public
License as published by the Free Software Foundation; either
version 2.1 of the License, or (at your option) any later version.

This library is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
Lesser General Public License for more details.

You should have received a copy of the GNU Lesser General Public
License along with this library; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
*/


#ifndef DEBUGCONSOLE_H
#define DEBUGCONSOLE_H

#include <QWidget>

namespace Ui {
    class DebugConsole;
}

class DebugConsole : public QWidget
{
    Q_OBJECT
public:
    explicit DebugConsole(QWidget *parent);
    ~DebugConsole();

    // Display debug message
    void appendDebugLogMessage(QString debugMessage);
    void appendDebugLockMessage(QString lockMessage);
    // Clear console
    void clearConsole();

private:
    // GUI components
    Ui::DebugConsole *ui;

private slots:
    // Perform a debug step (unlock)
    void step();

signals:
    // Step performed
    void stepped();
};

#endif // DEBUGCONSOLE_H
